#!/bin/csh

if (`flagLoc $argv -help`) then
   echo "Create a Psuedo-2D by Selecting a Series of 1D Spectral Regions:"
   echo " -in   inName   Input 1D Spectrum.           [test.ft1]"
   echo " -out  outName  Output 2D Spectrum.          [band.ft1]"
   echo " -mask maskName Output 2D Mask.              [mask.ft1]"
   echo " -tab  tabName  Output Table of Band Coords. [band.tab]"
   echo " -tmp  tmpName  Temporary Output File.       [tmp_mask_$$.ft1]"
   echo "Analysis Region (ppm Hz % pts):"
   echo " -x1  x1        [5%]       First Coord of Region to Analyze."
   echo " -xn  xn        [95%]      Last Coord of Region to Analyze."
   echo "Raised-Cosine Band Selection Window (ppm Hz % pts):"
   echo " -xw    xw      [0.6%]     +/- Window Plateau Width."
   echo " -xRamp xRamp   [2.0%]     Window Ramp Width."
   echo " -xStep xStep   [0.4%]     Window Position Step Size."
   exit 0
endif

set inName   = (`getArgD $argv -in     test.ft1`)
set outName  = (`getArgD $argv -out    band.ft1`)
set maskName = (`getArgD $argv -mask   mask.ft1`)
set tabName  = (`getArgD $argv -tab    band.tab`)
set tmpName  = (`getArgD $argv -tmp    tmp_mask_$$.ft1`)

set x1       = (`getArgD $argv -x1     5%`)
set xn       = (`getArgD $argv -xn     95%`)

set xw       = (`getArgD $argv -xw     0.6%`)
set xRamp    = (`getArgD $argv -xRamp  2.0%`)
set xStep    = (`getArgD $argv -xStep  0.4%`)

#
#-------------------------------------------------------------------------------------------

if (!(-e $inName)) then
   echo "bandSeries. Error finding 1D spectrum input $inName"
   exit 1
endif

/bin/rm -f $maskName $outName $tabName $tmpName

echo ""                                                              >  $tabName
echo "#"                                                             >> $tabName
echo "# Band coordinates:"                                           >> $tabName
echo "#  Left Ramp: X1 to X2 Plateau: X3 to X4 Right Ramp: X5 to X6" >> $tabName
echo ""                                                              >> $tabName
echo "VARS    INDEX BAND_X1 BAND_X2 BAND_X3 BAND_X4 BAND_X5 BAND_X6" >> $tabName
echo "FORMAT  %5d   %6d     %6d     %6d     %6d     %6d     %6d"     >> $tabName
echo ""                                                              >> $tabName

set ix1    = (`spec2pnt $inName CUR_XDIM $x1`)
set ixn    = (`spec2pnt $inName CUR_XDIM $xn`)

set ixw    = (`spec2pntW $inName CUR_XDIM $xw`)
set ixStep = (`spec2pntW $inName CUR_XDIM $xStep`)
set ixRamp = (`spec2pntW $inName CUR_XDIM $xRamp`)

set xSize  = (`getParm -in $inName -parm NDSIZE -dim CUR_XDIM -fmt %.0f`)

set ix1 = (`IMATH "integer($ix1)"`)
set ixn = (`IMATH "integer($ixn)"`)

if ($ix1 > $ixn) then
   set itmp = $ix1
   set ix1  = $ixn
   set ixn  = $itmp
endif

set x1PPM  = (`pnt2spec $inName CUR_XDIM $ix1 ppm`)
set xnPPM  = (`pnt2spec $inName CUR_XDIM $ixn ppm`)

set ixw    = (`IMATH "integer($ixw)"`)
set ixStep = (`IMATH "integer($ixStep)"`)
set ixRamp = (`IMATH "integer($ixRamp)"`)

nmrPrintf "Input Spectrum:     %s\n" $inName
nmrPrintf "Input PPM Range:    %.3f ppm to %.3f ppm, %s pts\n" $x1PPM $xnPPM $xSize
nmrPrintf "Analysis Region:    %s to %s (%s to %s pts)\n"  $x1 $xn $ix1 $ixn
nmrPrintf "Band Plateu Width:  +/- %s (+/- %s pts)\n" $xw $ixw
nmrPrintf "Band Ramp Width:    %s (%s pts)\n" $xRamp $ixRamp
nmrPrintf "Band Step:          %s (%s pts)\n" $xStep $ixStep
nmrPrintf "Output Spectrum:    %s\n" $outName
nmrPrintf "Output Mask:        %s\n" $maskName
nmrPrintf "Output Coord Table: %s\n" $tabName

set rampUpX1 = 1
set rampUpXN = (`IMATH "$rampUpX1 + $ixRamp - 1"`)

set platX1     = (`IMATH "$rampUpXN + 1"`)
set platXN     = (`IMATH "$platX1 + 2*$ixw"`)

set rampDownX1 = (`IMATH $platXN + 1`)
set rampDownXN = (`IMATH $rampDownX1 + $ixRamp - 1`)

normSpec.tcl -in $inName -out $tmpName -noverb \
 -prog set region 0%         100%       to 0.0 then \
       set region $rampUpX1 $rampDownXN to 1.0 then \
       rampUp   region $rampUpX1   $rampUpXN   then \
       rampDown region $rampDownX1 $rampDownXN

@ offset = (`IMATH "1 + $ix1 - ($ixRamp + integer($ixw/2))"`)
@ loc    = $ix1
@ n      = 1

nmrPipe -in $tmpName -fn CS -rs $offset -out $maskName -inPlace

@ rampUpX1   += $offset
@ rampUpXN   += $offset

@ platX1     += $offset
@ platXN     += $offset

@ rampDownX1 += $offset
@ rampDownXN += $offset

/bin/cp $inName $outName

nmrPrintf " %5d %6d %6d %6d %6d %6d %6d\n" $n $rampUpX1 $rampUpXN $platX1 $platXN $rampDownX1 $rampDownXN >> $tabName

while ($loc <= $ixn)
   @ offset += $ixStep
   @ loc += $ixStep
   @ n++

   nmrPipe -in $tmpName -fn CS   -rs $offset -tty -outHdr None >> $maskName
   nmrPipe -in $inName  -fn NULL             -tty -outHdr None >> $outName

   @ rampUpX1   += $ixStep
   @ rampUpXN   += $ixStep

   @ platX1     += $ixStep
   @ platXN     += $ixStep

   @ rampDownX1 += $ixStep
   @ rampDownXN += $ixStep

   nmrPrintf " %5d %6d %6d %6d %6d %6d %6d\n" $n $rampUpX1 $rampUpXN $platX1 $platXN $rampDownX1 $rampDownXN >> $tabName
end

sethdr $outName  -title Masked -ndim 2 -yN $n -yT $n -ySW $n -yOBS 1.0 -yCAR 0.0 -yMODE Real -yLAB Offset -yFT Time
sethdr $maskName -title Mask   -ndim 2 -yN $n -yT $n -ySW $n -yOBS 1.0 -yCAR 0.0 -yMODE Real -yLAB Offset -yFT Time 

addNMR -in1 $outName -in2 $maskName -out $outName -mult

nmrPrintf "Final Region Count: %s\n" $n

/bin/rm -f $tmpName

exit 0
