#!/bin/csh

if (`flagLoc $argv -help`) then
   echo "In-Place Processing of Spectral Series Listed in a Metadata Table."
   echo "Version 2025.087.16.17"
   echo " -meta      metaName [meta.tab]     Metadata Table (INDEX DIR_NAME TITLE)."
   echo " -fidName   fidName  [test.fid]     Input Name of FID In Its Source Directory."
   echo " -ftName    ftName   [Auto]         Output Name of Processed Spectrum In Its Original Directory."
   echo " -list      listName [ft.list]      Output List of Processed Spectra In Their Original Directories."
   echo " -ftDir     ftDir    [ft]           Optional Output Directory for Processed Spectra, or Keyword None."
   echo " -srcDir    srcDir   [.]            Source Directory."
   echo " -altScript sName    [None]         Alternative Processing Script for Special Cases."
   echo "Files Identifying Non-Uniformly Sampled Data (NUS):"
   echo " -mask    maskName [mask.fid]     NMRPipe-Format NUS Mask Name."
   echo " -sample  sName    [nuslist]      NUS Schedule Text File."
   echo "Processing Command; Should be Last on Command-Line:"
   echo " -cmnd    procCmnd [Auto]         Processing Command (Auto for basicFT1.com basicFT2.com ist2D.com)."
   echo "Notes:"
   echo "  1. The command specified by the -cmnd option should accept -in and -out options."
   echo "  2. If -ftName Auto is used, the output spectrum will be named test.ft1 or test.ft2 etc."
   echo "  3. If the -ftDir option is used, the output files will be named according to"
   echo "     the INDEX and TITLE from the metadata table."
   echo "  4. If the -altScript option is used, if the specified script exists in a given data directory,"
   echo "     it will be executed instead of using the processing command specified via the -cmnd option."
   echo "     This alternative script will be executed using all of the arguments that would be passed to"
   echo "     the processing command, and it should produce output files with the same names." 
   echo ""
   exit 0
endif

set procCmnd = "Auto"
set loc      = (`flagLoc $argv -cmnd`)
set thisArgv = ($argv)

if ($loc >= 1) then
   @ locP = $loc - 1
   @ locN = $loc + 1

   set procCmnd = ($argv[$locN-])
   set thisArgv = ($argv[1-$locP])
endif

set srcDir     = (`getArgD $thisArgv -srcDir    .`)
set ftDir      = (`getArgD $thisArgv -ftDir     ft`)
set fidName    = (`getArgD $thisArgv -fidName   test.fid`)
set ftName     = (`getArgD $thisArgv -ftName    Auto`)
set maskName   = (`getArgD $thisArgv -mask      mask.fid`)
set schedName  = (`getArgD $thisArgv -sample    nuslist`)
set tabName    = (`getArgD $thisArgv -meta      meta.tab`)
set listName   = (`getArgD $thisArgv -list      ft.list`)
set altScript  = (`getArgD $thisArgv -altScript None`)

#
#----------------------------------------------------------------------------------

set origDir = (`pwd`)

cd $srcDir
   set srcDirFull = (`pwd`)
cd $origDir

/bin/rm -f $listName

if (`string csame $ftDir None`) then
   set ftDirFull = None
else
   /bin/rm -rf $ftDir
   mkdir $ftDir

   cd $ftDir
      set ftDirFull = (`pwd`)
   cd $origDir
endif

set idList   = (`getTabCol.tcl -in $tabName -var INDEX`)
set dirList  = (`getTabCol.tcl -in $tabName -var DIR_NAME`)
set titList  = (`getTabCol.tcl -in $tabName -var TITLE`)

@ dimCount = 2
@ nusFlag  = 0

set thisName = $srcDir/$dirList[1]/$fidName

if (-e $thisName) then
   @ dimCount = (`getParm -in $thisName -parm FDDIMCOUNT -dim NULL_DIM -fmt %.0f`)

   if ((-e $srcDir/$dirList[1]/$maskName) || (-e $srcDir/$dirList[1]/$schedName)) then
      @ nusFlag = 1
   endif
endif

set thisExt = ft$dimCount

if (`string csame $ftName Auto`) then
   set ftName = test.${thisExt}
endif

if (`string csame $procCmnd Auto`) then
   if ($nusFlag) then
      set procCmnd = "ist2D.com"
   else
      set procCmnd = "basicFT${dimCount}.com"
   endif
endif

echo ""

@ eCount = 0
@ i = 0

foreach d ($dirList)
   @ i++

   if (!(-d $srcDirFull/$d)) then
      echo "Processing Error Finding Input Directory $srcDirFull/$d"
      @ eCount++
      continue
   endif

   cd $srcDirFull/$d
      set thisDir   = (`pwd`)
      set thisTitle = (`nmrPrintf "%03d-%s" $idList[$i] $titList[$i]`)

      if (!(-e $fidName)) then
         echo "Processing Error Finding FID Input $srcDirFull/$d/$fidName"
         @ eCount++
         continue
      endif

      if (`string csame $ftDir None`) then
         set outName = $srcDirFull/$d/$ftName
      else
         set outName = $ftDirFull/${thisTitle}.${thisExt}
      endif

      /bin/rm -f $ftName mask.fid fid.com proc.out

      if (`string notcsame $altScript None` && (-e $altScript)) then
         echo "Processing Spectrum $i of $#idList" Using Alternate Processing $altScript':'
         echo " Input Directory:  $thisDir"
         echo " Output Spectrum:  $outName"
         echo " Title:            $thisTitle"

         $altScript $procCmnd -in $fidName -out $ftName -title $thisTitle 
      else
         echo "Processing Spectrum $i of $#idList"':'
         echo " Input Directory:  $thisDir"
         echo " Output Spectrum:  $outName"
         echo " Title:            $thisTitle"

         $procCmnd -in $fidName -out $ftName -title $thisTitle 
      endif

      if (!(-e $ftName)) then
         echo "Processing Error Creating $srcDirFull/$d/$ftName"
         @ eCount++
         continue
      endif

      if (`string notcsame $ftDir None`) then
         /bin/cp $ftName $outName
      endif

      echo $thisDir/$ftName >> $origDir/$listName

      reportND.com $outName
      echo ""
   cd $origDir
end

if ($eCount != 0) then
   echo "Processing Notice: Errors Occurred During Processing."
   exit 1
endif

exit 0
