/***/
/* NMRPipe Copyright 1992 - 2024 Frank Delaglio
/*  simTimeND: simulation of N-dimensional complex time-domain data.
/*             Updated 7/2/2024 to correct decay formulas (useNewLB).
/***/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "cmndargs.h"
#include "dataio.h"
#include "memory.h"
#include "fdatap.h"
#include "dimloc.h"
#include "nmrio.h"
#include "table.h"
#include "rand.h"
#include "prec.h"
#include "vutil.h"

#define FPR  (void)fprintf
#define MAXJ 4

static int save1D(), saveHyperplane(), sim1D(), sim1DE(), sim1DG(), simPnt(), simPntE(), simPntG();
static int fixSimHdr(), showSim(), hyperProd(), getJList(), getJVals(), getparams();

static float findAmpE(), findAmpG();

static int useNewLB = 1;

/***/
/* Shorthand for error test of file write:
/***/

#define WRITE2D( M, Z, A ) \
  if ((error = saveHyperplane(&outUnit,M,outFmt,timeSize,dimCount,iz,ia,Z,A,fdata,fmtCount))) \
     { \
      (void) fprintf( stderr, "SimTimeND Error Writing data.\n" ); \
      goto shutdown; \
     } error = 0
 
int main( argc, argv )

   int  argc;
   char **argv;
{
    int     posLoc[MAXDIM], widthLoc[MAXDIM], timeSize[MAXDIM], freqSize[MAXDIM], ctSize[MAXDIM];
    int     cosJLoc[MAXDIM][MAXJ], sinJLoc[MAXDIM][MAXJ], p0Loc[MAXDIM], p1Loc[MAXDIM];
    int     negFlag[MAXDIM], altFlag[MAXDIM], gaussFlag[MAXDIM];

    float   p0Val[MAXDIM], p1Val[MAXDIM], thisP0Val[MAXDIM], thisP1Val[MAXDIM], offsetVal[MAXDIM];
    float   cosJVals[MAXJ], sinJVals[MAXJ];

    float   fdata[FDATASIZE], *matrix, *work1, *work2, *work3, *work4,
            *peakMatrix, *d1R, *d1I, *d2R, *d2I, *pkPtr, zR, zI, aR, aI,
            noise, ampScale, thisHi;

    int     i, j, iz, ia, n, aRealFlag, aRealLimit, peak, varCount, peakCount, planeCount, 
            dimCount, planesOut, totalPlanes, fmtCount, checkFlag, ovFlag, posCnt, widthCnt, 
            cosJCount, sinJCount, thisCosJCount, thisSinJCount, p0Count, p1Count, useTimeScale,
            useTDDecay, useTDJMod, hiLoc, scaleVarLoc, verbose, error;

    NMR_INT planeSize64;

    char    inName[NAMELEN+1], outName[NAMELEN+1], dirName[NAMELEN+1], outFmt[NAMELEN+1], axisOrder[NAMELEN+1];
    char    hiVarName[NAMELEN+1], scaleVarName[NAMELEN+1];
    struct  tabVar *varInfo;
    UNIT    outUnit;

/***/
/* Initialize:
/***/

    error       = 0;

    *outName    = '\0';
    *outFmt     = '\0';
    outUnit     = UNIT_NULL;

    matrix      = (float *) NULL;
    work1       = (float *) NULL;
    work2       = (float *) NULL;
    work3       = (float *) NULL;
    work4       = (float *) NULL;

    d1R         = (float *) NULL;
    d1I         = (float *) NULL;
    d2R         = (float *) NULL;
    d2I         = (float *) NULL;

    peakMatrix  = (float *) NULL;
    varInfo     = (struct tabVar *) NULL;

    planeSize64 = 0;
    peakCount   = 0;
    varCount    = 0;
    ampScale    = 1.0;

    useTimeScale = 1;
    useTDDecay   = 0;
    useTDJMod    = 1;

    thisCosJCount = 0;
    thisSinJCount = 0;

    hiLoc         = -1;
    scaleVarLoc   = -1;

    (void) initArgs( argc, argv, CHECK_ENV );

    (void) strcpy( inName,       "None" );
    (void) strcpy( hiVarName,    "HEIGHT" );
    (void) strcpy( scaleVarName, "None" );

    (void) iSet( timeSize,   MAXDIM,  1 );
    (void) iSet( freqSize,   MAXDIM,  1 );
    (void) iSet( posLoc,     MAXDIM, -1 );
    (void) iSet( widthLoc,   MAXDIM, -1 );
    (void) iSet( p0Loc,      MAXDIM, -1 );
    (void) iSet( p1Loc,      MAXDIM, -1 );
    (void) iSet( ctSize,     MAXDIM,  0 );
    (void) iSet( negFlag,    MAXDIM,  0 );
    (void) iSet( altFlag,    MAXDIM,  0 );
    (void) iSet( gaussFlag,  MAXDIM,  0 );

    (void) rSet( p0Val,      MAXDIM, (float)0.0 );
    (void) rSet( p1Val,      MAXDIM, (float)0.0 );

    (void) rSet( thisP0Val,  MAXDIM, (float)0.0 );
    (void) rSet( thisP1Val,  MAXDIM, (float)0.0 );

    (void) rSet( cosJVals,   MAXDIM, (float)0.0 );
    (void) rSet( sinJVals,   MAXDIM, (float)0.0 );

    (void) rSet( offsetVal,  MAXDIM, (float)0.0 );

    for( i = 0; i < MAXDIM; i++ )
       {
        for( j = 0; j < MAXJ; j++ )
           {
            cosJLoc[i][j] = -1;
            sinJLoc[i][j] = -1;
           }
       }

    (void) setHdrNull( fdata );

/***/
/* Extract command arguments.
/***/

    error = getparams( argc, argv, 
                       inName, outFmt, fdata, ctSize, hiVarName, scaleVarName,
                       negFlag, altFlag, gaussFlag, offsetVal, axisOrder, &dimCount, &noise, &ampScale,
                       &checkFlag, &ovFlag,  &fmtCount, &useTimeScale, &useTDDecay, &useTDJMod,
                       &verbose );

    if (error)
       {
        FPR( stderr, "SimTime Error extracting arguments.\n" );
        goto shutdown;
       }

    for( i = 0; i < dimCount; i++ )
       {
        thisP0Val[i] = p0Val[i] = getParm( fdata, NDP0, i+1 ); 
        thisP1Val[i] = p1Val[i] = getParm( fdata, NDP1, i+1 );
       }

    if ((error = fixSimHdr( argc, argv, timeSize, freqSize, fdata, fmtCount )))
       {
        FPR( stderr, "SimTime Error in header specification.\n" );
        goto shutdown;
       }

/***/
/* Read peak table if needed and extract parameter locations:
/***/

    if (*inName && strcasecmp( inName, "None" ))
       {
        if ((error = rdTable( inName, &peakMatrix, &varCount, &peakCount, &varInfo, -666 )))
           {
            FPR( stderr, "SimTime Error reading %s\n", inName );
            goto shutdown;
           }

        hiLoc = getTabVar( varInfo, varCount, hiVarName );

        if (hiLoc < 0)
           {
            FPR( stderr, "SimTime Error: missing height variable %s.\n", hiVarName );
            error = 1;
            goto shutdown;
           }

        if (strcasecmp( scaleVarName, "None" ))
           {
            scaleVarLoc = getTabVar( varInfo, varCount, scaleVarName );

            if (scaleVarLoc < 0)
               {
                FPR( stderr, "SimTime Error: missing height scale variable %s.\n", scaleVarName );
                error = 1;
                goto shutdown;
               }
           }

        posCnt   = getPosList( posLoc, MAXDIM, varInfo, varCount, axisOrder );
        widthCnt = getLWList( widthLoc, MAXDIM, varInfo, varCount, axisOrder );
  
        if (posCnt < dimCount || widthCnt < dimCount)
           {
            FPR( stderr, "SimTime Error: Insufficient Peak Table Information.\n" );
            error = 1;
            goto shutdown;
           }

        p0Count   = getParmList( p0Loc, MAXDIM, varInfo, varCount, axisOrder, "P0" );
        p1Count   = getParmList( p1Loc, MAXDIM, varInfo, varCount, axisOrder, "P1" );

        cosJCount = getJList( cosJLoc, MAXDIM, MAXJ, varInfo, varCount, axisOrder, "COSJ" );
        sinJCount = getJList( sinJLoc, MAXDIM, MAXJ, varInfo, varCount, axisOrder, "SINJ" );

        if (verbose)
           {
            FPR( stderr, "\n" );
            FPR( stderr, "SimTimeND Table Variables:\n" );
            FPR( stderr, " Height:       %s\n", hiVarName );
            FPR( stderr, " Height Scale: %s\n", scaleVarName );
            FPR( stderr, "\n" );
            FPR( stderr, " Position:      " );
            for( i = 0; i < dimCount; i++ ) FPR( stderr, " %-9s", (varInfo + posLoc[i])->name ); 
            FPR( stderr, "\n" );

            FPR( stderr, " Line Width:    " );
            for( i = 0; i < dimCount; i++ ) FPR( stderr, " %-9s", (varInfo + widthLoc[i])->name );
            FPR( stderr, "\n" );
            FPR( stderr, "\n" );

            FPR( stderr, " P0 Phase:      " );
            for( i = 0; i < dimCount; i++ ) FPR( stderr, " %-9s", p0Loc[i] < 0 ? "None" : (varInfo + p0Loc[i])->name );
            FPR( stderr, "\n" );

            FPR( stderr, " P1 Phase:      " );
            for( i = 0; i < dimCount; i++ ) FPR( stderr, " %-9s", p1Loc[i] < 0 ? "None" : (varInfo + p1Loc[i])->name );
            FPR( stderr, "\n" );

            for( j = 0; j < MAXJ; j++ )
               {
                n = 0;
                for( i = 0; i < dimCount; i++ ) if (cosJLoc[i][j] >= 0) n++;

                if (n || !j)
                   { 
                    FPR( stderr, " In-Phase J:    " );
                    for( i = 0; i < dimCount; i++ ) FPR( stderr, " %-9s", cosJLoc[i][j] < 0 ? "None" : (varInfo + cosJLoc[i][j])->name );
                    FPR( stderr, "\n" );
                   }

                if (cosJCount < 1) break;
               }

            for( j = 0; j < MAXJ; j++ )
               {
                n = 0;
                for( i = 0; i < dimCount; i++ ) if (sinJLoc[i][j] >= 0) n++;
    
                if (n || !j)
                   {
                    FPR( stderr, " Anti-Phase J:  " );
                    for( i = 0; i < dimCount; i++ ) FPR( stderr, " %-9s", sinJLoc[i][j] < 0 ? "None" : (varInfo + sinJLoc[i][j])->name );
                    FPR( stderr, "\n" );
                   }

                if (sinJCount < 1) break;
               }

            FPR( stderr, "\n" );
           }
       }

/***/
/* Create output directory if needed:
/***/

    if (*outFmt)
       {
        (void) sprintf( outName, outFmt, 1, 1 );

        if (!dirExists( outName, dirName ))
           {
            if (verbose)
               {
                FPR( stderr, "SimTimeND Notice: Creating directory %s\n", dirName );
               }

            if (dirCreate( dirName ))
               {
                FPR( stderr, "SimTimeND Warning: Error creating directory %s\n", dirName );
               }
           }
       }


/***/
/* Allocate time domain buffers:
/***/

    totalPlanes = 1;
    planesOut   = 1;

    MTEST( "SimTime", d1R = fltAlloc( "sim", timeSize[XLOC] ));
    MTEST( "SimTime", d1I = fltAlloc( "sim", timeSize[XLOC] ));

    if (dimCount == 1)
       {
        planeSize64 = timeSize[XLOC];

        MTEST( "SimTime", work1  = fltAlloc( "sim", planeSize64 ));
        MTEST( "SimTime", work2  = fltAlloc( "sim", planeSize64 ));
       }
    else
       {
        planeSize64 = (NMR_INT)timeSize[XLOC]*(NMR_INT)timeSize[YLOC]*(NMR_INT)4;

        MTEST( "SimTime", matrix = fltAlloc( "sim", planeSize64 ));
        MTEST( "SimTime", work1  = fltAlloc( "sim", planeSize64 ));

        MTEST( "SimTime", d2R = fltAlloc( "sim", timeSize[YLOC] ));
        MTEST( "SimTime", d2I = fltAlloc( "sim", timeSize[YLOC] ));
       }

    if (dimCount >= 3)
       {
        MTEST( "SimTime", work2 = fltAlloc( "sim", planeSize64 ));
        totalPlanes *= timeSize[ZLOC]*2;
        planesOut   *= 2;
       }

    if (dimCount >= 4)
       {
        MTEST( "SimTime", work3 = fltAlloc( "sim", planeSize64 ));
        MTEST( "SimTime", work4 = fltAlloc( "sim", planeSize64 ));
        totalPlanes *= timeSize[ALOC]*2;
        planesOut   *= 2;
       }
 
/***/
/* For each A/Z plane in 2D series:
/*
/*   Initialize signal summation workspaces.
/*   Create workspaces.
/*
/*     For each peak:
/*       Simulate real and imaginary 1D profile for X dimension.
/*
/*       For higher dim peaks:
/*         Simulate real and imaginary 1D profile for Y dimension.
/*         Create products RR, IR, RI, II in the 2D data matrix.
/*
/*       For 1D and 2D Spectra
/*         Add the simulated signal to sum workspaces.
/*
/*       For 3D and 4D Spectra  
/*         Find Z or A dimension time-domain point value.
/*         Add Z/A modulated signal to sum workspaces.
/*
/*   Add noise to workspaces if needed. 
/*   Write the summations.
/***/

    planeCount = 1;
    aRealLimit = dimCount == 4 ? 0 : 1;

    if (ampScale == 0.0)
       {
        ampScale = 1.0;

        for( i = 0; i < dimCount; i++ ) ampScale /= (float)timeSize[i];
       }

    for( ia = 0; ia < timeSize[ALOC]; ia++ )
     {
      for( aRealFlag = 1; aRealFlag >= aRealLimit; aRealFlag-- )
       {    
        for( iz = 0; iz < timeSize[ZLOC]; iz++ )
           {
            switch( dimCount )
               {
                case 1:
                   (void) vsSet( work1, planeSize64, (float)0.0 );
                   (void) vsSet( work2, planeSize64, (float)0.0 );
                   break;
                case 2:
                   (void) vsSet( work1, planeSize64, (float)0.0 );
                   break;
                case 3:
                   (void) vsSet( work1, planeSize64, (float)0.0 );
                   (void) vsSet( work2, planeSize64, (float)0.0 );
                   break;
                case 4:
                   (void) vsSet( work1, planeSize64, (float)0.0 );
                   (void) vsSet( work2, planeSize64, (float)0.0 );
                   break;
                default:
                   break;
               }

            pkPtr = peakMatrix;
        
            for( peak = 1; peak <= peakCount; peak++ )
               {
                (void) showSim( iz, ia, aRealFlag, timeSize[ZLOC], timeSize[ALOC], dimCount, peak, peakCount, verbose );

                if (p0Count || p1Count)
                   {
                    for( i = 0; i < dimCount; i++ )
                       {
                        thisP0Val[i] = p0Loc[i] < 0 ? p0Val[i] : p0Val[i] + pkPtr[p0Loc[i]];
                        thisP1Val[i] = p1Loc[i] < 0 ? p1Val[i] : p1Val[i] + pkPtr[p1Loc[i]];
                       }
                  }


                if (cosJCount) thisCosJCount = getJVals( pkPtr, cosJLoc, cosJVals, XLOC );
                if (sinJCount) thisSinJCount = getJVals( pkPtr, sinJLoc, sinJVals, XLOC );

                thisHi = scaleVarLoc < 0 ? ampScale*pkPtr[hiLoc] : ampScale*pkPtr[hiLoc]*pkPtr[scaleVarLoc];

                (void) sim1D( d1R, d1I, 
                              timeSize[XLOC], freqSize[XLOC], ctSize[XLOC], gaussFlag[XLOC],
                              (float)(offsetVal[XLOC] + pkPtr[posLoc[XLOC]]), pkPtr[widthLoc[XLOC]],
                              cosJVals, thisCosJCount, sinJVals, thisSinJCount,
                              thisHi, thisP0Val[XLOC], thisP1Val[XLOC],
                              useTimeScale, useTDDecay, useTDJMod );

                if (negFlag[XLOC]) (void) vNeg( d1I, timeSize[XLOC] );
                if (altFlag[XLOC]) (void) vAlt( d1R, timeSize[XLOC] );
                if (altFlag[XLOC]) (void) vAlt( d1I, timeSize[XLOC] );
       
                if (dimCount > 1)
                   { 
                    if (cosJCount) thisCosJCount = getJVals( pkPtr, cosJLoc, cosJVals, YLOC );
                    if (sinJCount) thisSinJCount = getJVals( pkPtr, sinJLoc, sinJVals, YLOC );
 
                    (void) sim1D( d2R, d2I, 
                                  timeSize[YLOC], freqSize[YLOC], ctSize[YLOC], gaussFlag[YLOC],
                                  (float)(offsetVal[YLOC] + pkPtr[posLoc[YLOC]]), pkPtr[widthLoc[YLOC]],
                                  cosJVals, thisCosJCount, sinJVals, thisSinJCount,
                                  (float)1.0, thisP0Val[YLOC], thisP1Val[YLOC],
                                  useTimeScale, useTDDecay, useTDJMod );

                    if (negFlag[YLOC]) (void) vNeg( d2I, timeSize[YLOC] );
                    if (altFlag[YLOC]) (void) vAlt( d2R, timeSize[YLOC] );
                    if (altFlag[YLOC]) (void) vAlt( d2I, timeSize[YLOC] );
           
                    (void) hyperProd( matrix, d1R, d1I, d2R, d2I, timeSize[XLOC], timeSize[YLOC] );
                   }

                switch( dimCount )
                   { 
                    case 1:
                       (void) vvAdd( work1, d1R, planeSize64 );
                       (void) vvAdd( work2, d1I, planeSize64 );
                       break;

                    case 2:
                       (void) vvAdd( work1, matrix, planeSize64 );
                       break;

                    case 3:
                       if (cosJCount) thisCosJCount = getJVals( pkPtr, cosJLoc, cosJVals, ZLOC );
                       if (sinJCount) thisSinJCount = getJVals( pkPtr, sinJLoc, sinJVals, ZLOC );

                       (void) simPnt( &zR, &zI,
                                      timeSize[ZLOC], freqSize[ZLOC], ctSize[ZLOC], gaussFlag[ZLOC], iz,
                                      (float)(offsetVal[ZLOC] + pkPtr[posLoc[ZLOC]]), pkPtr[widthLoc[ZLOC]],
                                      cosJVals, thisCosJCount, sinJVals, thisSinJCount,
                                      (float)1.0, thisP0Val[ZLOC], thisP1Val[ZLOC], 
                                      useTimeScale, useTDDecay, useTDJMod );

                       if (negFlag[ZLOC]) zI = -zI;

                       if (altFlag[ZLOC] && (iz % 2))
                          {
                           zR = -zR;
                           zI = -zI;
                          }

                       (void) vsvAdd( work1, matrix, planeSize64, zR );
                       (void) vsvAdd( work2, matrix, planeSize64, zI );

                       break;

                    case 4:
                       if (cosJCount) thisCosJCount = getJVals( pkPtr, cosJLoc, cosJVals, ZLOC );
                       if (sinJCount) thisSinJCount = getJVals( pkPtr, sinJLoc, sinJVals, ZLOC );

                       (void) simPnt( &zR, &zI,
                                      timeSize[ZLOC], freqSize[ZLOC], ctSize[ZLOC], gaussFlag[ZLOC], iz,
                                      (float)(offsetVal[ZLOC] + pkPtr[posLoc[ZLOC]]), pkPtr[widthLoc[ZLOC]],
                                      cosJVals, thisCosJCount, sinJVals, thisSinJCount,
                                      (float)1.0, thisP0Val[ZLOC], thisP1Val[ZLOC],
                                      useTimeScale, useTDDecay, useTDJMod );

                       (void) simPnt( &aR, &aI,
                                      timeSize[ALOC], freqSize[ALOC], ctSize[ALOC], gaussFlag[ALOC], ia,
                                      (float)(offsetVal[ALOC] + pkPtr[posLoc[ALOC]]), pkPtr[widthLoc[ALOC]],
                                      cosJVals, thisCosJCount, sinJVals, thisSinJCount,
                                      (float)1.0, thisP0Val[ALOC], thisP1Val[ALOC], 
                                      useTimeScale, useTDDecay, useTDJMod );

                       if (negFlag[ZLOC]) zI = -zI;
                       if (negFlag[ALOC]) aI = -aI;

                       if (altFlag[ZLOC] && (iz % 2))
                          {
                           zR = -zR;
                           zI = -zI;
                          }

                       if (altFlag[ALOC] && (ia % 2))
                          {
                           aR = -aR;
                           aI = -aI;
                          }

                       if (cosJCount) thisCosJCount = getJVals( pkPtr, cosJLoc, cosJVals, ALOC );
                       if (sinJCount) thisSinJCount = getJVals( pkPtr, sinJLoc, sinJVals, ALOC );

                       if (aRealFlag)
                          {
                           (void) vsvAdd( work1, matrix, planeSize64, (float)(zR*aR) );
                           (void) vsvAdd( work2, matrix, planeSize64, (float)(zI*aR) );
                          }
                       else
                          {
                           (void) vsvAdd( work1, matrix, planeSize64, (float)(zR*aI) );
                           (void) vsvAdd( work2, matrix, planeSize64, (float)(zI*aI) );
                          }

                       break;

                    default:
                       break;
                   }

                pkPtr += varCount;
               } 

            switch( dimCount )
               {
                case 1:
                   (void) vAddGNoise( work1, planeSize64, noise );
                   (void) vAddGNoise( work2, planeSize64, noise );

                   error = save1D( outName, work1, work2, fdata, planeSize64 );
                   break;

                case 2:
                   (void) vAddGNoise( work1, planeSize64, noise );

                   WRITE2D( work1, 1, 1 );

                   planeCount += planesOut;
                   break;

                case 3:
                   (void) vAddGNoise( work1, planeSize64, noise );
                   (void) vAddGNoise( work2, planeSize64, noise );

                   WRITE2D( work1, 1, 1 );
                   WRITE2D( work2, 0, 1 );

                   planeCount += planesOut;
                   break;

                case 4:
                   (void) vAddGNoise( work1, planeSize64, noise );
                   (void) vAddGNoise( work2, planeSize64, noise );

                   WRITE2D( work1, 1, aRealFlag );
                   WRITE2D( work2, 0, aRealFlag );
 
                   planeCount += planesOut;
                   break;

                default:
                   break;
               }
           }
        }
     }    

    if (verbose)
       {
        FPR( stderr, "\n" );
        (void) fflush( stderr );
       }

/***/
/* Exit point:
/***/

shutdown:

    if (*outFmt) (void) dataClose( outUnit );

    (void) deAlloc( "sim", matrix, sizeof(float)*planeSize64 );

    (void) deAlloc( "sim", work1,  sizeof(float)*planeSize64 );
    (void) deAlloc( "sim", work2,  sizeof(float)*planeSize64 );
    (void) deAlloc( "sim", work3,  sizeof(float)*planeSize64 );
    (void) deAlloc( "sim", work4,  sizeof(float)*planeSize64 );

    (void) deAlloc( "sim", d1R, sizeof(float)*timeSize[XLOC] );
    (void) deAlloc( "sim", d1I, sizeof(float)*timeSize[XLOC] );
    (void) deAlloc( "sim", d2R, sizeof(float)*timeSize[YLOC] );
    (void) deAlloc( "sim", d2I, sizeof(float)*timeSize[YLOC] );

    (void) freeTab( peakMatrix, varInfo, varCount, peakCount );

    return( error );
}

/***/
/* Display current status of the simulation.
/***/

static int showSim( iz, ia, aRealFlag, zSize, aSize, dimCount, peak, peakCount, verbose )

   int iz, ia, aRealFlag, zSize, aSize, dimCount, peak, peakCount, verbose;
{
    if (verbose)
       {
        if (dimCount == 1)
           {
            FPR( stderr, "SimTimeND Peak %d of %d\015", peak, peakCount );
           }
        else if (dimCount == 2)
           {
            if (peak == 1 || peak == peakCount || !(peak % 16))
               {
                FPR( stderr, "SimTimeND Peak %d of %d\015", peak, peakCount );
               }
           }
        else if (dimCount == 3)
           {
            if (peak == 1 || peak == peakCount || !(peak % 16))
               {
                FPR( stderr, "SimTimeND Z: %3d of %d Peak: %4d of %d\015", iz + 1, zSize, peak, peakCount );
               }
           }
        else
           {
            if (peak == 1 || peak == peakCount || !(peak % 16))
               {
                FPR( stderr, 
                     "SimTimeND Z: %3d of %d A: %3d of %d %s\015", 
                     iz + 1, zSize, ia + 1, aSize, aRealFlag ? "Real" : "Imag" );
               }
           }

        (void) fflush( stderr );
       }

    return( 0 );
}

/***/
/* save1D: save a 1D complex FID.
/***/

static int save1D( outName, rdata, idata, fdata, tSize )

   NMR_INT tSize;
   char    *outName;
   float   *rdata, *idata, fdata[FDATASIZE]; 
{
    int  error;
    UNIT outUnit;

    error   = 0;
    outUnit = FB_STDOUT;

    if (*outName)
       {
        if ((error = dataOpen( outName, &outUnit, FB_CREATE )))
           {
            FPR( stderr, "SimTime Error opening output %s\n", outName );
            goto shutdown;
           }
       }

    (void) getNMRMinMax64( rdata, tSize, (NMR_INT)1, 1, fdata );

    if ((error = dataWrite( outUnit, fdata, sizeof(float)*FDATASIZE )))
       {
        FPR( stderr, "SimTime Error writing header to %s\n", outName );
        goto shutdown;
       }

    if ((error = dataWrite( outUnit, rdata, sizeof(float)*tSize )))
       {
        FPR( stderr, "SimTime Error writing real data to %s\n", outName );
        goto shutdown;
       }

    if ((error = dataWrite( outUnit, idata, sizeof(float)*tSize )))
       {
        FPR( stderr, "SimTime Error writing imag data to %s\n", outName );
        goto shutdown;
       }

shutdown:
 
    if (*outName) (void) dataClose( outUnit );

    return( error );
}

/***/
/* saveHyperplane: save a 2D hypercomplex plane from the current simulation.
/***/

static int saveHyperplane( outUnitPtr, matrix,
                           outFmt, sizeList, dimCount,
                           iz, ia, zReal, aReal, fdata, fmtCount )

    int   sizeList[MAXDIM], dimCount, iz, ia, zReal, aReal, fmtCount;
    float *matrix, fdata[FDATASIZE];
    UNIT  *outUnitPtr;
    char  *outFmt;
{
    char    outName[NAMELEN+1];
    int     zPlane, aPlane, createFlag, quadState, error;
    NMR_INT count64, offset64;

    static int  firstTime = 1;

/***/   
/* Find the plane number according to Z and A coords.
/*
/* Build the file name:
/*   A single file can be 1D-4D.
/*   A one-integer template can be 3D (2D plane series) or 4D (3D cube series).
/*   A two-integer template is always 4D (2D plane series).
/***/

    outName[0] = '\0';
    error      = 0;

    count64  = (NMR_INT)sizeList[XLOC]*(NMR_INT)sizeList[YLOC]*(NMR_INT)4;
    offset64 = (NMR_INT)0;

    if (zReal)
       zPlane = 2*iz + 1;
    else
       zPlane = 2*iz + 2;

    if (aReal)
       aPlane = 2*ia + 1;
    else
       aPlane = 2*ia + 2;

    if (fmtCount == 0)
       (void) strcpy( outName, outFmt );
    else if (fmtCount == 1)
       (void) sprintf( outName, outFmt, dimCount == 4 ? aPlane : zPlane );
    else
       (void) sprintf( outName, outFmt, aPlane, zPlane );

/***/
/* Decide whether to create the output file.
/* If outName is empty, data will be written to stdout.
/***/

    createFlag = 0;
    offset64   = (NMR_INT)0;

    if (firstTime) createFlag = 1;

    if (fmtCount == 0)
       {
        if (dimCount <= 2)
           {
            createFlag = 1;
           }
        else if (dimCount == 3)
           {
            if (zPlane == 1) createFlag = 1;
            offset64 = FDATASIZE + count64*(zPlane - 1);
           }
        else
           {
            if (zPlane == 1 && aPlane == 1) createFlag = 1;
            offset64 = FDATASIZE + count64*(zPlane - 1) + count64*2*sizeList[ZLOC]*(aPlane - 1);
           }
       }
    else if (fmtCount == 1)
       {
        if (dimCount == 4)
           {
            offset64 = FDATASIZE + count64*(zPlane - 1);
            if (zPlane == 1) createFlag = 1;
           }
        else
           {
            createFlag = 1;
           }
       }
    else if (fmtCount == 2)
       {
        createFlag = 1;
       }

    if (*outName)
       {
        if (createFlag)
           {
            if ((error = dataOpen( outName, outUnitPtr, FB_CREATE )))
               {
                FPR( stderr, "SimTime Error creating output %s\n", outName );
                goto shutdown;
               }
           }
        else
           {
            if ((error = dataOpen( outName, outUnitPtr, FB_WRITE )))
               {
                FPR( stderr, "SimTime Error opening output %s\n", outName );
                goto shutdown;
               }
           }
       }
    else
       {
        *outUnitPtr = FB_STDOUT;
       }

/***/
/* Update the display scale.
/* Write the header if needed.
/* Write the data.
/* Close the file if needed.
/***/

    quadState = getQuad( fdata, FDQUADFLAG, 0 );

    (void) getNMRMinMax64( matrix, (NMR_INT)sizeList[XLOC], (NMR_INT)sizeList[YLOC], quadState, fdata );

    if (createFlag)
       {
        if ((error = dataWrite( *outUnitPtr, fdata, sizeof(float)*FDATASIZE )))
           {
            FPR( stderr, "SimTime Error writing header to %s\n", outName );
            goto shutdown;
           }
      }

   if (*outName && offset64)
      {
       if ((error = dataPos( *outUnitPtr, sizeof(float)*offset64 )))
          {
           FPR( stderr, "SimTime Error positioning in file %s\n", outName );
           goto shutdown;
          }
      }

   if ((error = dataWrite( *outUnitPtr, matrix, sizeof(float)*count64 )))
      {
       FPR( stderr, "SimTime Error writing data to %s\n", outName );
       goto shutdown;
      }

/***/
/* Exit point:
/***/

shutdown:

    if (*outName)
       {
        (void) dataClose( *outUnitPtr );
        *outUnitPtr = UNIT_NULL;
       }

    firstTime = 0;

    return( error );
}

/***/
/* hyperProd: make a hypercomplex 2D product from complex 1D profiles.
/***/

static int hyperProd( matrix, d1R, d1I, d2R, d2I, xtSize, ytSize )

   float *matrix, *d1R, *d1I, *d2R, *d2I;
   int   xtSize, ytSize;
{
    float *d1RPtr, *d1IPtr, *d2RPtr, *d2IPtr;
    float *rrPtr, *riPtr, *irPtr, *iiPtr;
    int   ix, iy, jump;

    jump = 3*xtSize;

    rrPtr  = matrix;
    irPtr  = rrPtr + xtSize;
    riPtr  = irPtr + xtSize;
    iiPtr  = riPtr + xtSize;

    d2RPtr = d2R;
    d2IPtr = d2I;

    for( iy = 0; iy < ytSize; iy++ )
       {
        d1RPtr = d1R; d1IPtr = d1I;

        for( ix = 0; ix < xtSize; ix++ )
           {
            *rrPtr++ = (*d1RPtr)*(*d2RPtr);
            *irPtr++ = (*d1IPtr)*(*d2RPtr);
            *riPtr++ = (*d1RPtr)*(*d2IPtr);
            *iiPtr++ = (*d1IPtr)*(*d2IPtr);

            d1RPtr++; d1IPtr++;
           }

        rrPtr += jump; irPtr += jump; riPtr += jump; iiPtr += jump;

        d2RPtr++; d2IPtr++;
       }

    return( 0 );
}

/***/
/* getJVals: extract non-zero couplings for the given peak.
/***/
 
static int getJVals( pkPtr, jLoc, jVals, dim )

   float *pkPtr, jVals[MAXJ];
   int   jLoc[MAXDIM][MAXJ], dim;
{
    int j, n;

    n = 0;

    for( j = 0; j < MAXJ; j++ )
       {
        if (jLoc[dim][j] < 0) continue;
        if (pkPtr[jLoc[dim][j]] == 0.0) continue;
        jVals[n++] = pkPtr[jLoc[dim][j]];
       }

    return( n );
}

/***/
/* simPnt: simulate a single complex point from a damped sinusoid.
/***/

static int simPnt( rdata, idata, tsize, fsize, ctsize, gaussFlag, ix, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod )

   float *rdata, *idata, *cosJVals, *sinJVals, x0, xfw, hi, p0, p1;
   int   tsize, fsize, ctsize, gaussFlag, cosJCount, sinJCount, ix, useTimeScale, useTDDecay, useTDJMod;
{
    int status;

    if (gaussFlag)
       {
        status = simPntG( rdata, idata, tsize, fsize, ctsize, ix, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod );
       }
    else
       {
        status = simPntE( rdata, idata, tsize, fsize, ctsize, ix, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod );
       }

    return( status );
}

/***/
/* simPntE: simulate a single complex point from a exponentially-damped sinusoid.
/***/

static int simPntE( rdata, idata, tsize, fsize, ctsize, ix, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod )
  
   float *rdata, *idata, *cosJVals, *sinJVals, x0, xfw, hi, p0, p1;
   int   tsize, fsize, ctsize, cosJCount, sinJCount, ix, useTimeScale, useTDDecay, useTDJMod;
{
    float freq, f, tR, tI, q, r, lb, amp, phi, delay;
    int   j;

/***/
/* Freq coord x0 is interpreted in range of 1 to fsize.
/* Zero frequency in points 1->fsize is 1 + fsize/2.
/* Create point from time domain data of length tsize.
/***/

    if (xfw < 0.0) xfw = -xfw;

    if (ctsize < 0)     ctsize = 0;
    if (ctsize > tsize) ctsize = tsize;

    freq  = 2.0*PI*(x0 - (1 + fsize/2))/(float)fsize;

    if (useNewLB)
       lb = useTDDecay ? -0.5*xfw*PI/tsize : -xfw*PI/(float)fsize;
    else
       lb = useTDDecay ? -xfw/tsize : -xfw*PI/(float)fsize;

    delay = p1/360.0;
    phi   = PI*(p0 + p1/2.0)/180.0;
    f     = phi + (delay + ix)*freq;

    if (useTimeScale)
       {
        amp = findAmpE( tsize, fsize, ctsize, xfw, hi, useTDDecay )*exp( (double) ix*lb );
       }
    else
       {
        if (ctsize)
           amp =  ix < ctsize ? hi : hi*exp( (double)(1 + ix - ctsize)*lb );
        else
           amp =  hi*exp( (double)ix*lb );
       }

    *rdata  =  amp*cos( (double)f );
    *idata  = -amp*sin( (double)f );

    for( j = 0; j < cosJCount; j++ )
       {
        q = useTDJMod ? PI*cosJVals[j]/(tsize - 1) : PI*cosJVals[j]/(fsize - 1);
        r = cos( (double)(q*ix) );

        *rdata *= r;
        *idata *= r;
       }

    for( j = 0; j < sinJCount; j++ )
       {
        q = useTDJMod ? PI*sinJVals[j]/(tsize - 1) : PI*sinJVals[j]/(fsize - 1);
        r = sin( (double)(q*ix) );

        tR = *rdata;
        tI = *idata;
        
        *rdata =  tI*r;
        *idata = -tR*r;
       }

    return( 0 );
}

/***/
/* simPntG: simulate a single complex point from a Gaussian-damped sinusoid.
/***/

static int simPntG( rdata, idata, tsize, fsize, ctsize, ix, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod )
  
   float *rdata, *idata, *cosJVals, *sinJVals, x0, xfw, hi, p0, p1;
   int   tsize, fsize, ctsize, cosJCount, sinJCount, ix, useTimeScale, useTDDecay, useTDJMod;
{
    float freq, f, tR, tI, q, r, lb, amp, phi, delay;
    int   j;

/***/
/* Freq coord x0 is interpreted in range of 1 to fsize.
/* Zero frequency in points 1->fsize is 1 + fsize/2.
/* Create point from time domain data of length tsize.
/***/

/***/
/* Magic number is sqrt(2/(8*log(2))) = 0.6005612
/***/

    if (xfw < 0.0) xfw = -xfw;

    if (ctsize < 0)     ctsize = 0;
    if (ctsize > tsize) ctsize = tsize;

    freq  = 2.0*PI*(x0 - (1 + fsize/2))/(float)fsize;

    if (useNewLB)
       lb = useTDDecay ? 0.5*0.6005612*xfw*PI/tsize : 0.6005612*xfw*PI/fsize;
    else
       lb = useTDDecay ?     0.6005612*xfw*PI/tsize : 0.6005612*xfw*PI/fsize;

    delay = p1/360.0;
    phi   = PI*(p0 + p1/2.0)/180.0;
    f     = phi + (delay + ix)*freq;

    if (useTimeScale)
       {
        amp = findAmpG( tsize, fsize, ctsize, xfw, hi, useTDDecay )*exp( -(double)((ix*lb)*(ix*lb)) );
       }
    else
       {
        if (ctsize)
           amp =  ix < ctsize ? hi : hi*exp( -(double)( ((1 + ix - ctsize)*lb)*((1 + ix - ctsize)*lb) ) );
        else
           amp =  hi*exp( -(double)((ix*lb)*(ix*lb)) );
       }

    *rdata  =  amp*cos( (double)f );
    *idata  = -amp*sin( (double)f );

    for( j = 0; j < cosJCount; j++ )
       {
        q = useTDJMod ? PI*cosJVals[j]/(tsize - 1) : PI*cosJVals[j]/(fsize - 1);
        r = cos( (double)(q*ix) );

        *rdata *= r;
        *idata *= r;
       }

    for( j = 0; j < sinJCount; j++ )
       {
        q = useTDJMod ? PI*sinJVals[j]/(tsize - 1) : PI*sinJVals[j]/(fsize - 1);
        r = sin( (double)(q*ix) );

        tR = *rdata;
        tI = *idata;
        
        *rdata =  tI*r;
        *idata = -tR*r;
       }

    return( 0 );
}

/***/
/* findAmpG: return sinusoid amplitude adjusted for Gaussian decay profile.
/***/

static float findAmpG( tsize, fsize, ctsize, xfw, hi, useTDDecay )
  
   int   tsize, fsize, ctsize, useTDDecay;
   float xfw, hi;
{
    float lb, sum;
    int   ix;

/***/
/* Magic number is sqrt(2/(8*log(2))) = 0.6005612
/***/

    if (xfw < 0.0) xfw = -xfw;

    if (ctsize < 0)     ctsize = 0;
    if (ctsize > tsize) ctsize = tsize;

    if (useNewLB)
       lb = useTDDecay ? 0.5*0.6005612*xfw*PI/tsize : 0.6005612*xfw*PI/fsize;
    else
       lb = useTDDecay ?     0.6005612*xfw*PI/tsize : 0.6005612*xfw*PI/fsize;

    sum   = 0.0;

    if (ctsize)
       {
        sum += (float)ctsize;

        for( ix = ctsize; ix < tsize; ix++ )
           {
            sum += exp( -(double)(((1 + ix - ctsize)*lb)*((1 + ix - ctsize)*lb)) );
           }
       }
    else
       {
        for( ix = 0; ix < tsize; ix++ )
           {
            sum += exp( -(double)((ix*lb)*(ix*lb)) );
           }
       }

    sum /= (float)tsize;

    if (sum != 0.0) hi /= sum;

    return( hi );
}

/***/
/* findAmpE: return sinusoid amplitude adjusted for exponential decay profile.
/***/

static float findAmpE( tsize, fsize, ctsize, xfw, hi, useTDDecay )
  
   int   tsize, fsize, ctsize, useTDDecay;
   float xfw, hi;
{
    float lb, sum;
    int   ix;

    if (xfw < 0.0) xfw = -xfw;

    if (ctsize < 0)     ctsize = 0;
    if (ctsize > tsize) ctsize = tsize;

    if (useNewLB)
       lb = useTDDecay ? -0.5*xfw*PI/tsize : -xfw*PI/(float)fsize;
    else
       lb = useTDDecay ? -xfw/tsize : -xfw*PI/(float)fsize;

    sum   = 0.0;

    if (ctsize)
       {
        sum += (float)ctsize;

        for( ix = ctsize; ix < tsize; ix++ )
           {
            sum += exp( (double)(1 + ix - ctsize)*lb );
           }
       }
    else
       {
        for( ix = 0; ix < tsize; ix++ )
           {
            sum += exp( (double)ix*lb );
           }
       }

    sum /= (float)tsize;

    if (sum != 0.0) hi /= sum;

    return( hi );
}

/***/
/* sim1D: simulate a 1D complex damped sinusoid.
/***/

static int sim1D( rdata, idata, tsize, fsize, ctsize, gaussFlag, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod )

   float *rdata, *idata, *cosJVals, *sinJVals, x0, xfw, hi, p0, p1;
   int   tsize, fsize, ctsize, gaussFlag, cosJCount, sinJCount, useTimeScale, useTDDecay, useTDJMod;
{
    int status;

    if (gaussFlag)
       {
        status = sim1DG( rdata, idata, tsize, fsize, ctsize, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod );
       }
    else
       {
        status = sim1DE( rdata, idata, tsize, fsize, ctsize, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod );
       }

    return( status );
}

/***/
/* sim1DE: simulate a 1D complex exponentially-damped sinusoid.
/***/

static int sim1DE( rdata, idata, tsize, fsize, ctsize, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod )
   
   float *rdata, *idata, *cosJVals, *sinJVals, x0, xfw, hi, p0, p1;
   int   tsize, fsize, ctsize, cosJCount, sinJCount, useTimeScale, useTDDecay, useTDJMod;
{
    float *rPtr, *iPtr, tR, tI, f, q, freq, lb, sum, amp, phi, delay;
    int   ix, j;

/***/
/* Freq coord x0 is interpreted in range of 1 to fsize.
/* Zero frequency in points 1->fsize is 1 + fsize/2.
/* Create time domain data of length tsize.
/* Compute total intensity of decay function.
/***/

    if (xfw < 0.0) xfw = -xfw;

    if (ctsize < 0)     ctsize = 0;
    if (ctsize > tsize) ctsize = tsize;

    freq  = 2.0*PI*(x0 - (1 + fsize/2))/(float)fsize;

    if (useNewLB)
       lb = useTDDecay ? -0.5*xfw*PI/tsize : -xfw*PI/(float)fsize;
    else
       lb = useTDDecay ? -xfw/tsize : -xfw*PI/(float)fsize;

    sum   = 0.0;

    delay = p1/360.0;
    phi   = PI*(p0 + p1/2.0)/180.0;

    rPtr  = rdata;
    iPtr  = idata;

    if (ctsize)
       {
        for( ix = 0; ix < ctsize; ix++ )
           {
            amp      = 1.0;
            f        =  phi + (delay + ix)*freq;
            *rPtr++  =  hi*amp*cos( (double)f );
            *iPtr++  = -hi*amp*sin( (double)f );
            sum     +=  amp;
           }

        for( ix = ctsize; ix < tsize; ix++ )
           {
            amp      =  exp( (double)(1 + ix - ctsize)*lb );
            f        =  phi + (delay + ix)*freq;
            *rPtr++  =  hi*amp*cos( (double)f );
            *iPtr++  = -hi*amp*sin( (double)f );
            sum     +=  amp;
           }
       }
    else
       {
        for( ix = 0; ix < tsize; ix++ )
           {
            amp      =  exp( (double)ix*lb );
            f        =  phi + (delay + ix)*freq;
            *rPtr++  =  hi*amp*cos( (double)f );
            *iPtr++  = -hi*amp*sin( (double)f );
            sum     +=  amp;
           }
       }

/***/
/* Apply couplings if any:
/***/

   for( j = 0; j < cosJCount; j++ )
      {
       rPtr  = rdata;
       iPtr  = idata;

       q = useTDJMod ? PI*cosJVals[j]/(tsize - 1) : PI*cosJVals[j]/(fsize - 1);

       for( ix = 0; ix < tsize; ix++ )
          {
           amp = cos( (double)(q*ix) );
           *rPtr++ *= amp;
           *iPtr++ *= amp;
          }
      }

   for( j = 0; j < sinJCount; j++ )
      {
       rPtr  = rdata;
       iPtr  = idata;

       q = useTDJMod ? PI*sinJVals[j]/(tsize - 1) : PI*sinJVals[j]/(fsize - 1);

       for( ix = 0; ix < tsize; ix++ )
          {
           amp = sin( (double)(q*ix) );

           tR = *rPtr;
           tI = *iPtr;

           *rPtr++ =  tI*amp;
           *iPtr++ = -tR*amp;
          }
      }

/***/
/* Scale the FID by decay-sum to give desired height after FT:
/***/

    sum /= (float)tsize;

    if (!useTimeScale) return( 0 );
    if (sum == 0.0)    return( 0 );
 
    amp = 1.0/sum;

    rPtr = rdata;
    iPtr = idata;

    for( ix = 0; ix < tsize; ix++ )
       {
        *rPtr++ *= amp;
        *iPtr++ *= amp;
       }

    return( 0 );
}

/***/
/* sim1DG: simulate a 1D complex Gaussian-damped sinusoid.
/***/

static int sim1DG( rdata, idata, tsize, fsize, ctsize, x0, xfw, cosJVals, cosJCount, sinJVals, sinJCount, hi, p0, p1, useTimeScale, useTDDecay, useTDJMod )
   
   float *rdata, *idata, *cosJVals, *sinJVals, x0, xfw, hi, p0, p1;
   int   tsize, fsize, ctsize, cosJCount, sinJCount, useTimeScale, useTDDecay, useTDJMod;
{
    float *rPtr, *iPtr, tR, tI, f, q, freq, lb, sum, amp, phi, delay;
    int   ix, j;

/***/
/* Freq coord x0 is interpreted in range of 1 to fsize.
/* Zero frequency in points 1->fsize is 1 + fsize/2.
/* Create time domain data of length tsize.
/* Compute total intensity of decay function.
/***/

/***/
/* Magic number is sqrt(2/(8*log(2))) = 0.6005612
/***/

    if (xfw < 0.0) xfw = -xfw;

    if (ctsize < 0)     ctsize = 0;
    if (ctsize > tsize) ctsize = tsize;

    freq  = 2.0*PI*(x0 - (1 + fsize/2))/(float)fsize;

    if (useNewLB)
       lb = useTDDecay ? 0.5*0.6005612*xfw*PI/tsize : 0.6005612*xfw*PI/fsize;
    else
       lb = useTDDecay ?     0.6005612*xfw*PI/tsize : 0.6005612*xfw*PI/fsize;

    sum   = 0.0;

    delay = p1/360.0;
    phi   = PI*(p0 + p1/2.0)/180.0;

    rPtr  = rdata;
    iPtr  = idata;

    if (ctsize)
       {
        for( ix = 0; ix < ctsize; ix++ )
           {
            amp      = 1.0;
            f        =  phi + (delay + ix)*freq;
            *rPtr++  =  hi*amp*cos( (double)f );
            *iPtr++  = -hi*amp*sin( (double)f );
            sum     +=  amp;
           }

        for( ix = ctsize; ix < tsize; ix++ )
           {
            amp      =  exp( -(double)(((1 + ix - ctsize)*lb)*((1 + ix - ctsize)*lb)) );
            f        =  phi + (delay + ix)*freq;
            *rPtr++  =  hi*amp*cos( (double)f );
            *iPtr++  = -hi*amp*sin( (double)f );
            sum     +=  amp;
           }
       }
    else
       {
        for( ix = 0; ix < tsize; ix++ )
           {
            amp      =  exp( -(double)((ix*lb)*(ix*lb)) );
            f        =  phi + (delay + ix)*freq;
            *rPtr++  =  hi*amp*cos( (double)f );
            *iPtr++  = -hi*amp*sin( (double)f );
            sum     +=  amp;
           }
       }

/***/
/* Apply couplings if any:
/***/

   for( j = 0; j < cosJCount; j++ )
      {
       rPtr  = rdata;
       iPtr  = idata;

       q = useTDJMod ? PI*cosJVals[j]/(tsize - 1) : PI*cosJVals[j]/(fsize - 1);

       for( ix = 0; ix < tsize; ix++ )
          {
           amp = cos( (double)(q*ix) );
           *rPtr++ *= amp;
           *iPtr++ *= amp;
          }
      }

   for( j = 0; j < sinJCount; j++ )
      {
       rPtr  = rdata;
       iPtr  = idata;

       q = useTDJMod ? PI*sinJVals[j]/(tsize - 1) : PI*sinJVals[j]/(fsize - 1);

       for( ix = 0; ix < tsize; ix++ )
          {
           amp = sin( (double)(q*ix) );

           tR = *rPtr;
           tI = *iPtr;

           *rPtr++ =  tI*amp;
           *iPtr++ = -tR*amp;
          }
      }

/***/
/* Scale the FID by decay-sum to give desired height after FT:
/***/

    sum /= (float)tsize;

    if (!useTimeScale) return( 0 );
    if (sum == 0.0)    return( 0 );
 
    amp = 1.0/sum;

    rPtr = rdata;
    iPtr = idata;

    for( ix = 0; ix < tsize; ix++ )
       {
        *rPtr++ *= amp;
        *iPtr++ *= amp;
       }

    return( 0 );
}

/***/
/* fixSimHdr: create and adjust header for simulation output.
/***/

static int fixSimHdr( argc, argv, timeSize, freqSize, fdata, fmtCount )

   int   argc, timeSize[MAXDIM], freqSize[MAXDIM], fmtCount;
   float fdata[FDATASIZE];
   char  **argv;
{
    int   dim, dimCount, fileCount, pipeFlag, mid;
    float orig, sw, car, obs, value;

/***/
/* For each axis:
/*   Test for hypercomplex data.
/*   Extract axis time and frequency size.
/*   Replace frequency size with time size. 
/*   Update chemical shift information.
/***/

    if (flagLoc( argc, argv, "-xMODE" ) && !flagLoc( argc, argv, "-aqMODE" ))
       {
        value = getParm( fdata, NDQUADFLAG, CUR_XDIM );
        (void) setParm( fdata, FDQUADFLAG, value, 0 );
       }

    if (flagLoc( argc, argv, "-xT" ) && !flagLoc( argc, argv, "-aqN" ))
       {
        value = getParm( fdata, NDAPOD, CUR_XDIM );
        (void) setParm( fdata, FDREALSIZE, value, 0 );
       }

    dimCount = getParm( fdata, FDDIMCOUNT, 0 );

    for( dim = 1; dim <= dimCount; dim++ )
       {
        (void) setParm( fdata, NDP0, (float)0.0, dim );
        (void) setParm( fdata, NDP1, (float)0.0, dim );

        if (1 == getQuad( fdata, NDQUADFLAG, dim ))
           {
            FPR( stderr, "SimTime Error: Non-Complex Axis %d.\n", dim );
            return( 1 );
           }

        timeSize[dim-1]  = getParm( fdata, NDAPOD, dim );
        freqSize[dim-1]  = getParm( fdata, NDSIZE, dim );
        car              = getParm( fdata, NDCAR, dim );
        obs              = getParm( fdata, NDOBS, dim );
        sw               = getParm( fdata, NDSW,  dim );

        mid              = 1 + timeSize[dim-1]/2;
        orig             = obs*car - sw*(timeSize[dim-1] - mid)/timeSize[dim-1];

        (void) setParm( fdata, NDTDSIZE, (float)timeSize[dim-1], dim );

        switch( dim )
           {
            case CUR_XDIM:
               if (!flagLoc( argc, argv, "-xORIG" ))
                  {
                   (void) setParm( fdata, NDORIG, orig, dim );
                  }

               (void) setParm( fdata, NDSIZE,   (float)timeSize[dim-1],   dim );
               (void) setParm( fdata, NDCENTER, (float)mid,               dim );
               break;

            case CUR_YDIM:
               if (!flagLoc( argc, argv, "-xORIG" ))
                  {
                   (void) setParm( fdata, NDORIG, orig, dim );
                  }

               (void) setParm( fdata, NDSIZE,   (float)(2*timeSize[dim-1]), dim );
               (void) setParm( fdata, NDCENTER, (float)mid,                 dim );
               break;

            case CUR_ZDIM:
               if (!flagLoc( argc, argv, "-xORIG" ))
                  {
                   (void) setParm( fdata, NDORIG, orig, dim );
                  }

               (void) setParm( fdata, NDSIZE,   (float)(2*timeSize[dim-1]), dim );
               (void) setParm( fdata, NDCENTER, (float)mid,                 dim );
               break;

            case CUR_ADIM:
               if (!flagLoc( argc, argv, "-xORIG" ))
                  {
                   (void) setParm( fdata, NDORIG, orig, dim );
                  }

               (void) setParm( fdata, NDSIZE,   (float)(2*timeSize[dim-1]), dim );
               (void) setParm( fdata, NDCENTER, (float)mid,                 dim );
               break;

            default:
               (void) setParm( fdata, NDSIZE,   (float)(2*timeSize[dim-1]), dim );
               (void) setParm( fdata, NDCENTER, (float)mid,                 dim );
               (void) setParm( fdata, NDORIG,   orig,                       dim );
               break;
           }
       }

/***/
/* Update chemical shift information:
/***/

/***/
/* Adjust header according to single or multi-file mode.
/* Parse explicit user header definitions.
/***/

    fileCount = 1;
    pipeFlag  = 0;

    if (fmtCount == 0)
       {
        if (dimCount > 2)
           {
            pipeFlag = CUR_XDIM;
           }
       }
    else if (fmtCount == 1)
       {
        if (dimCount == 3)
           fileCount  = timeSize[ZLOC]*2;
        else
           fileCount  = timeSize[ZLOC]*timeSize[ALOC]*4;
       }
    else
       {
        fileCount  = timeSize[ZLOC]*timeSize[ALOC]*4;
       }

    (void) setParm( fdata, FDFILECOUNT, (float)fileCount, 0 );
    (void) setParm( fdata, FDPIPEFLAG,  (float)pipeFlag,  0 );

    if (fmtCount == 1 && dimCount == 4)
       (void) setParm( fdata, FDCUBEFLAG, (float)1, 0 );
    else
       (void) setParm( fdata, FDCUBEFLAG, (float)0, 0 );
       
    if (parseHdr2( argc, argv, fdata )) return( 1 );

    return( 0 );
}

/***/
/* Extract command arguments for simTimeND
/***/

static int getparams( argc, argv,
                      inName, outFmt, fdata, ctSize, hiVarName, scaleVarName,
                      negFlag, altFlag, gaussFlag, offsetVal, axisOrder, dimCount, noise, ampScale,
                      checkFlag, ovFlag, fmtCount, useTimeScale, useTDDecay, useTDJMod, verbose )

   int   argc, *dimCount, *fmtCount, *checkFlag, *ovFlag, *useTimeScale, *useTDDecay, *useTDJMod, *verbose;
   int   ctSize[MAXDIM], negFlag[MAXDIM], altFlag[MAXDIM], gaussFlag[MAXDIM];

   char  **argv, *inName, *outFmt, *axisOrder, *hiVarName, *scaleVarName;
   float fdata[FDATASIZE], offsetVal[MAXDIM], *noise, *ampScale;
{
    char ctemp[NAMELEN+1];
    int  iseed;

/***/
/* Initialize:
/***/

    iseed         = 54321;

    *verbose      = 0;
    *useTimeScale = 1;
    *useTDDecay   = 0;
    *useTDJMod    = 1;
    *ovFlag       = 0;
    *checkFlag    = 0;
    *fmtCount     = 0;
    *dimCount     = 2;
    *ampScale     = 1.0;

    *outFmt       = '\0';
    *noise        = 0.0;

    (void) strcpy( axisOrder, "XYZABC" );

/***/
/* Print a banner:
/***/

    if (argc == 1 || flagLoc( argc, argv, "-help" ))
       {
        FPR( stderr, "SimTimeND: ND Complex Time-Domain Simulation.\n" ); 
        FPR( stderr, "%s hdrFlags > pipeFile\n", argv[0] );
        FPR( stderr, "SimTimeND-Specific Flags:\n" );
        FPR( stderr, " -in       inName  [None]   Peak Table Input, or Keyword None.\n" );
        FPR( stderr, " -out      outName [stdout] NMRPipe Format Output, or Keyword stdout.\n" );
        FPR( stderr, " -hiVar    hiVar   [HEIGHT] Table Variable for Peak Height.\n" );
        FPR( stderr, " -scaleVar sVar    [None]   Table Variable for Height Scale.\n" );
        FPR( stderr, " -scale    s       [1.0]    Amplituide Scale Factor (Zero for Auto).\n" );
        FPR( stderr, " -axis     ord     [XYZA]   Axis Order List.\n" );
        FPR( stderr, " -rms      nVal    [0.0]    RMS of Noise to Add (Also: -noise).\n" );
        FPR( stderr, " -iseed    iseed   [54321]  Random Number Seed for Noise.\n" );
        FPR( stderr, " -ts                        Scale Time-Domain by Decay Integral (Default).\n" );
        FPR( stderr, " -nots                      No Time-Domain Scale.\n" );
        FPR( stderr, " -notdd                     Interpret Linewidth in Frequency Domain (Default).\n" );
        FPR( stderr, " -tdd                       Interpret Linewidth as Time Domain Decay.\n" );
        FPR( stderr, " -tdj                       Interpret J-Modulation in Time Domain (Default).\n" );
        FPR( stderr, " -notdj                     Interpret J-Modulation in Frequency Domain.\n" );
        FPR( stderr, " -useOld                    Use Legacy Formulas for Decay.\n" );
        FPR( stderr, " -ov                        Force Overwrite.\n" );
        FPR( stderr, " -nofs                      No File Size Check.\n" );
        FPR( stderr, " -verb                      Verbose Mode ON.\n" );
        FPR( stderr, " -noverb                    Verbose Mode OFF (Default).\n" );
        FPR( stderr, "Exponential or Gaussian Decay (Use -x -y -z -a):\n" );
        FPR( stderr, " -xExp                      Use Exponential Decay (Default).\n" );
        FPR( stderr, " -xGauss                    Use Gaussian Decay.\n" );
        FPR( stderr, "Offset in Frequency Domain Points (Use -x -y -z -a):\n" );
        FPR( stderr, " -xOff     xOff [0.0]       X-Axis Peak Position Offset.\n" );
        FPR( stderr, "Sign Adjustment (Use -x -y -z -a):\n" );
        FPR( stderr, " -xNeg                      Negate Imaginary Data.\n" );
        FPR( stderr, " -xAlt                      Negate Alternate Points.\n" );
        FPR( stderr, "Alternative Specification of Sign Adjustment (Use -x -y -z -a):\n" );
        FPR( stderr, " -xFTARG   arg  [None]      Keyword: None, neg alt neg,alt.\n" );
        FPR( stderr, "Signals with Initial Constant Time Period (Use -x -y -z -a):\n" );
        FPR( stderr, " -xCT      xCT  [0]         Number of Constant-Time Points in X-Axis.\n" );
        FPR( stderr, "\n" );

        (void) parseHdr( argc, argv, fdata );

        FPR( stderr, "Notes:\n" );
        FPR( stderr, " 1. Simulation freq-domain sizes -xN -yN etc should correspond to\n" );
        FPR( stderr, "    the size of the all-real spectrum that input table is based upon.\n" );
        FPR( stderr, " 2. All dimensions in the output will be complex.\n" );
        FPR( stderr, " 3. The input table requires HEIGHT, a point position\n" );
        FPR( stderr, "    and linewidth in points for each dimension (X_AXIS XW ...)\n" );
        FPR( stderr, " 4. If the input table contains XTW YTW ... these will be\n" );
        FPR( stderr, "    used instead of XW YW etc. This allows differentiation of\n" );
        FPR( stderr, "    time-domain decays and observed frequency-domain widths.\n" );
        FPR( stderr, " 5. The global phase of all signals is specified by command line\n" );
        FPR( stderr, "    options -xP0 -xP1 etc. If the input table has columns X_P0 X_P1\n" );
        FPR( stderr, "    etc, these column values will be added to the signal's phase.\n" );
        FPR( stderr, " 6. If the input table contains X_COSJ1 ... X_COSJ4 Y_COSJ1 etc.,\n" );
        FPR( stderr, "    with values in time-domain points, cosine modulation will\n" );
        FPR( stderr, "    be applied. Likewise for X_SINJ... etc, sine modulation will\n" );
        FPR( stderr, "    be applied. Using -notdj will interpret modulation in freq-\n" );
        FPR( stderr, "    domain points instead.\n" );
        FPR( stderr, " 7. Any output, 1D-4D, can be sent to a single file or pipeline.\n" );
        FPR( stderr, " 8. 3D output can also be to single value template such as:\n" );
        FPR( stderr, "     ft/test%%03d.fid      (2D Plane Series)\n" );
        FPR( stderr, " 9. 4D output can also be to single or double value template:\n" );
        FPR( stderr, "     ft/test%%02d%%03d.fid  (2D Plane Series).\n" );
        FPR( stderr, "     ft/test%%03d.fid      (3D Cube Series).\n" );

        FPR( stderr, "\n" );
        return( 1 );
      }

/***/
/* Extract command-specific parameters:
/***/

    (void) intArgD( argc, argv, "-ndim",  dimCount );
    (void) intArgD( argc, argv, "-iseed", &iseed );

    (void) intArgD( argc, argv, "-xCT", ctSize + XLOC );
    (void) intArgD( argc, argv, "-yCT", ctSize + YLOC );
    (void) intArgD( argc, argv, "-zCT", ctSize + ZLOC );
    (void) intArgD( argc, argv, "-aCT", ctSize + ALOC );

    (void) fltArgD( argc, argv, "-xOff", offsetVal + XLOC );
    (void) fltArgD( argc, argv, "-yOff", offsetVal + YLOC );
    (void) fltArgD( argc, argv, "-zOff", offsetVal + ZLOC );
    (void) fltArgD( argc, argv, "-aOff", offsetVal + ALOC );

    if (offsetVal[XLOC] != 0.0) offsetVal[XLOC] -= 1.0;
    if (offsetVal[YLOC] != 0.0) offsetVal[YLOC] -= 1.0;
    if (offsetVal[ZLOC] != 0.0) offsetVal[ZLOC] -= 1.0;
    if (offsetVal[ALOC] != 0.0) offsetVal[ALOC] -= 1.0;

    (void) strcpy( ctemp, "None" );
    (void) strArgD( argc, argv, "-xFTARG", ctemp );

    if (strstr( ctemp, "neg" )) negFlag[XLOC] = 1;
    if (strstr( ctemp, "alt" )) altFlag[XLOC] = 1;

    (void) strcpy( ctemp, "None" );
    (void) strArgD( argc, argv, "-yFTARG", ctemp );

    if (strstr( ctemp, "neg" )) negFlag[YLOC] = 1;
    if (strstr( ctemp, "alt" )) altFlag[YLOC] = 1;

    (void) strcpy( ctemp, "None" );
    (void) strArgD( argc, argv, "-zFTARG", ctemp );

    if (strstr( ctemp, "neg" )) negFlag[ZLOC] = 1;
    if (strstr( ctemp, "alt" )) altFlag[ZLOC] = 1;

    (void) strcpy( ctemp, "None" );
    (void) strArgD( argc, argv, "-aFTARG", ctemp );

    if (strstr( ctemp, "neg" )) negFlag[ALOC] = 1;
    if (strstr( ctemp, "alt" )) altFlag[ALOC] = 1;

    if (flagLoc( argc, argv, "-xNeg" )) negFlag[XLOC] = 1;
    if (flagLoc( argc, argv, "-yNeg" )) negFlag[YLOC] = 1;
    if (flagLoc( argc, argv, "-zNeg" )) negFlag[ZLOC] = 1;
    if (flagLoc( argc, argv, "-aNeg" )) negFlag[ALOC] = 1;

    if (flagLoc( argc, argv, "-xAlt" )) altFlag[XLOC] = 1;
    if (flagLoc( argc, argv, "-yAlt" )) altFlag[YLOC] = 1;
    if (flagLoc( argc, argv, "-zAlt" )) altFlag[ZLOC] = 1;
    if (flagLoc( argc, argv, "-aAlt" )) altFlag[ALOC] = 1;

    if (flagLoc( argc, argv, "-xExp" )) gaussFlag[XLOC] = 0;
    if (flagLoc( argc, argv, "-yExp" )) gaussFlag[YLOC] = 0;
    if (flagLoc( argc, argv, "-zExp" )) gaussFlag[ZLOC] = 0;
    if (flagLoc( argc, argv, "-aExp" )) gaussFlag[ALOC] = 0;

    if (flagLoc( argc, argv, "-xGauss" )) gaussFlag[XLOC] = 1;
    if (flagLoc( argc, argv, "-yGauss" )) gaussFlag[YLOC] = 1;
    if (flagLoc( argc, argv, "-zGauss" )) gaussFlag[ZLOC] = 1;
    if (flagLoc( argc, argv, "-aGauss" )) gaussFlag[ALOC] = 1;

    if (flagLoc(  argc, argv, "-useOld" ))   useNewLB = 0;
    if (flagLoc(  argc, argv, "-nouseOld" )) useNewLB = 1;
    if (flagLoc(  argc, argv, "-useNew" ))   useNewLB = 1;

    SRAND( iseed );

    (void) strArgD( argc, argv, "-in",       inName );
    (void) strArgD( argc, argv, "-out",      outFmt );
    (void) strArgD( argc, argv, "-axis",     axisOrder );
    (void) strArgD( argc, argv, "-hiVar",    hiVarName );
    (void) strArgD( argc, argv, "-scaleVar", scaleVarName );
    (void) fltArgD( argc, argv, "-rms",      noise );
    (void) fltArgD( argc, argv, "-noise",    noise );
    (void) fltArgD( argc, argv, "-scale",    ampScale );

    if (flagLoc( argc, argv, "-tdd" ))    *useTDDecay   = 1;
    if (flagLoc( argc, argv, "-notdd" ))  *useTDDecay   = 0;

    if (flagLoc( argc, argv, "-tdj" ))    *useTDJMod    = 1;
    if (flagLoc( argc, argv, "-notdj" ))  *useTDJMod    = 0;

    if (flagLoc( argc, argv, "-ts" ))     *useTimeScale = 1;
    if (flagLoc( argc, argv, "-nots" ))   *useTimeScale = 0;

    if (flagLoc( argc, argv, "-nofs" ))   *checkFlag    = 0;
    if (flagLoc( argc, argv, "-ov" ))     *ovFlag       = 1;

    if (flagLoc( argc, argv, "-verb" ))   *verbose      = 1;
    if (flagLoc( argc, argv, "-noverb" )) *verbose      = 0;

    if (!strcasecmp( outFmt, "stdout" )) *outFmt = '\0';
    if (!strcmp( outFmt, "-" ))          *outFmt = '\0';

    *fmtCount = getFmtCount( outFmt );

    if (parseHdr( argc, argv, fdata )) return( 1 );

    (void) fixfdata( fdata );

    return( 0 );
}

int getPosList( posList, maxDim, varInfo, varCount, axisOrder )

   int    *posList, maxDim, varCount;
   struct tabVar *varInfo;
   char   *axisOrder;
{
    static char axisLabel[] = "X_AXIS";
    int    dim, posCount;

    posCount = 0;

    for( dim = 0; dim < maxDim; dim++ )
       {
        axisLabel[0] = axisOrder[dim]; 
        *posList     = getTabVar( varInfo, varCount, axisLabel );

        if (*posList < 0) return( posCount );

        posCount++;
        posList++;
       }

    return( posCount );
} 

int getParmList( parmList, maxDim, varInfo, varCount, axisOrder, prefix )

   int    parmList[MAXDIM], maxDim, varCount;
   char   *axisOrder, *prefix;
   struct tabVar *varInfo;
{
    char parmLabel[NAMELEN+1];
    int  dim, pCount;

    pCount = 0;

    for( dim = 0; dim < maxDim; dim++ )
       {
        (void) sprintf( parmLabel, "%c_%s", axisOrder[dim], prefix );

        parmList[dim] = getTabVar( varInfo, varCount, parmLabel );

        if (parmList[dim] < 0) return( pCount );

        pCount++;
       }

    return( pCount );
}

int getLWList( lwList, maxDim, varInfo, varCount, axisOrder )

   int    *lwList, maxDim, varCount;
   struct tabVar *varInfo;
   char   *axisOrder;
{
    static char axisLabelT[] = "XTW";
    static char axisLabel[]  = "XW";
    int    dim, lwCount;

    lwCount = 0;

    for( dim = 0; dim < maxDim; dim++ )
       {
        axisLabelT[0] = axisOrder[dim];
        axisLabel[0]  = axisOrder[dim];

        *lwList = getTabVar( varInfo, varCount, axisLabelT );

        if (*lwList < 0)
           {
            *lwList = getTabVar( varInfo, varCount, axisLabel );
           }

        if (*lwList < 0) return( lwCount );

        lwCount++;
        lwList++;
       }

    return( lwCount );
}

static int getJList( jList, maxDim, maxJ, varInfo, varCount, axisOrder, prefix )

   int    jList[MAXDIM][MAXJ], maxDim, maxJ, varCount;
   char   *axisOrder, *prefix;
   struct tabVar *varInfo;
{
    char   jLabel[NAMELEN+1];
    int    j, dim, jCount;

    jCount = 0;

    for( dim = 0; dim < maxDim; dim++ )
       {
        for( j = 0; j < maxJ; j++ )
           {
            (void) sprintf( jLabel, "%c_%s%d", axisOrder[dim], prefix, j+1 );

            jList[dim][j] = getTabVar( varInfo, varCount, jLabel );

            if (jList[dim][j] >= 0) jCount++; 
           }
       }

    return( jCount );
}

/***/
/* Bottom.
/***/

